##
# geometry.itcl
#
# Routines to assist in manipulating screen co-ordinates
# in various ways. There are several groups of functions, to deal
# with screen geometry, points, vertices, and rectangles.
#
# NOTE: These are stil being developed, and the interfaces are
# likely to change.
#


##########################################################################
#
# Author:  H. John Reekie
#
# Version: @(#)geometry.tcl	1.14    12/16/96
#
# Copyright (c) 1990-1996 The Regents of the University of California.
# All rights reserved.
# 
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
# 
# 						PT_COPYRIGHT_VERSION_2
# 						COPYRIGHTENDKEY
##########################################################################


##########################################################################
#### degrees r
#
# Convert radians to degrees.
#
proc degrees {r} {
    global pi

    return [expr $r * 180.0 / $pi]
}

##########################################################################
#### radians r
#
# Convert degrees to radians.
#
proc radians {d} {
    global pi

    return [expr $d / 180.0 * $pi]
}

##########################################################################
#
## Point functions
#
# These functions assist in dealing with co-ordinates expressed as
# ``points'' where a point is a pair (that is, a 2-list) of numbers.
# Note that most of these routines expect each point to be a distinct
# list, not concatenated into the argument list. For example,
#
#    pointAdd {1 2} {3 4}
#
# is correct, but
#
#    pointAdd 1 2 3 4
#
# is not.
#

##########################################################################
#### pointx, pointy
#
# Extract coordinates from a point
#
proc pointx {point} {
    return [lindex $point 0]
}

proc pointy {point} {
    return [lindex $point 1]
}

##########################################################################
#### pointIncrement point x y
#
# Add coordinates to a point.
#
proc pointIncrement {point x y} {
    return [list \
	    [expr [pointx $point] + $x] \
	    [expr [pointy $point] + $y] ]
}
  
##########################################################################
#### pointDecrement point x y
#
# Subtract coordinates from a point
#
proc pointDecrement {point x y} {
    return [list \
	    [expr [pointx $point] - $x] \
	    [expr [pointy $point] - $y] ]
}

##########################################################################
#### pointAdd point offset
#
# Add two points
#
proc pointAdd {point offset} {
    return [list \
	    [expr [lindex $point 0] + [lindex $offset 0]] \
	    [expr [lindex $point 1] + [lindex $offset 1]] ]
}

##########################################################################
#### pointSubtract point offset
#
# Subtract two points. That is, return _point_ - _offset_.
#
proc pointSubtract {point offset} {
    return [list \
	    [expr [pointx $point] - [pointx $offset]] \
	    [expr [pointy $point] - [pointy $offset]] ]
}

##########################################################################
#### pointScale point xscale {yscale}
#
# Scale a point. If only the first scale argument is given
# scale the point equally in both directions -- that is,
# given the point (x, y), return (x * xscale, y * xscale). If
# both scale arguments are given, scale the point by those amounts
# in the x and y directions.
#
proc pointScale {point xscale args} {
    if { $args == "" } {
	set yscale $xscale
    } else {
	set yscale [lindex $args 0]
    }

    return [list \
	    [expr [lindex $point 0] * $xscale] \
	    [expr [lindex $point 1] * $yscale] ]
}

##########################################################################
#### pointAxpy a x y
#
# Scale a point and then add it to another. In other words, the result
# is ax + y.
#
proc pointAxpy {a x y} {
    return [list \
	    [expr [lindex $x 0] * $a + [lindex $y 0]] \
	    [expr [lindex $x 1] * $a + [lindex $y 1]]]
}

##########################################################################
#### pointVxpy v x y
#
# Scale a point by a vector and then add it to another. In other
# words, the result is vx + y.
#
proc pointVxpy {v x y} {
    return [list \
	    [expr [lindex $x 0] * [lindex $v 0] + [lindex $y 0]] \
	    [expr [lindex $x 1] * [lindex $v 1] + [lindex $y 1]]]
}

##########################################################################
#### pointFlip point
#
# ``Flip'' a point about the origin. In other words, negate its
# two components.
#
proc pointFlip {point} {
    return [list \
	    [expr -[pointx $point]] \
	    [expr -[pointy $point]] ]
}

##########################################################################
#
## Rectangle functions
#
# These functions assist in dealing with co-ordinates expressed as
# ``rectangles'' where a rectangle is a 4-list of numbers. As for
# points, rectangles are distinct lists, not concatenated into the
# argument list.
#

##########################################################################
#### rectx0, rectx1, recty0, rexty1
#
# Extract co-ordinates from a rectangle
#
proc rectx0 {rectangle} {
    return [lindex $rectangle 0]
}

proc recty0 {rectangle} {
    return [lindex $rectangle 1]
}

proc rectx1 {rectangle} {
    return [lindex $rectangle 2]
}

proc recty1 {rectangle} {
    return [lindex $rectangle 3]
}

##########################################################################
#### rectSize rect
#
# Return the size of a rectangle
#
proc rectSize {rect} {
    return [list \
	    [expr [rectx1 $rect] - [rectx0 $rect]] \
	    [expr [recty1 $rect] - [recty0 $rect]]]
}

##########################################################################
#### rectWidth rect
#
# Return the width of a rectangle
#
proc rectWidth {rect} {
    return [expr [rectx1 $rect] - [rectx0 $rect]]
}

##########################################################################
#### rectHeight rect
#
# Return the height of a rectangle
#
proc rectHeight {rect} {
    return [expr [recty1 $rect] - [recty0 $rect]]
}

##########################################################################
#### rectStrut rect
#
# Return the strut of a rectangle (as an interval). That is, return
# the two y-coordinates of the top and bottom edges of the rectangle.
#
proc rectStrut {rect} {
    return [list [recty1 $rect] [recty0 $rect]]
}

##########################################################################
#### rectSpan rect
#
# Return the span of a rectangle (as an interval). That is, return
# the two x-coordinates of the left and right edges of the rectangle.
#
proc rectSpan {rect} {
    return [list [rectx0 $rect] [rectx1 $rect]]
}

##########################################################################
#### rectStretch rect hdelta {vdelta}
#
# Stretch a rectangle on all sides. The rectangle will be larger
# in the horizontal direction by 2 * _hdelta_ pixels, and by
# 2 * _vdelta_ pixels in the vertical direction. If _vdelta_
# is omitted, then _hdelta_ is used in both directions.
#
proc rectStretch {rect hdelta args} {
    if { $args == "" } {
	set vdelta $hdelta
    } else {
	set vdelta [lindex $args 0]
    }

    return [list \
	    [expr [rectx0 $rect] - $hdelta] \
	    [expr [recty0 $rect] - $vdelta] \
	    [expr [rectx1 $rect] + $hdelta] \
	    [expr [recty1 $rect] + $vdelta] ]
}

##########################################################################
#### rectShrink rect hdelta {vdelta}
#
# Shrink a rectangle on all sides. The rectangle will be smaller
# in the horizontal direction by 2 * _hdelta_ pixels, and by
# 2 * _vdelta_ pixels in the vertical direction. If _vdelta_
# is omitted, then _hdelta_ is used in both directions.
#
proc rectShrink {rect hdelta args} {
    if { $args == "" } {
	set vdelta $hdelta
    } else {
	set vdelta [lindex $args 0]
    }

    return [list \
	    [expr [lindex $rect 0] + $hdelta] \
	    [expr [lindex $rect 1] + $vdelta] \
	    [expr [lindex $rect 2] - $hdelta] \
	    [expr [lindex $rect 3] - $vdelta] ]
}

##########################################################################
#### rectAdd rect point
#
# Move a rectangle by the distance specified by the given point.
#
proc rectAdd {rect point} {
    return [list \
	    [expr [rectx0 $rect] + [pointx $point]] \
	    [expr [recty0 $rect] + [pointy $point]] \
	    [expr [rectx1 $rect] + [pointx $point]] \
	    [expr [recty1 $rect] + [pointy $point]] ]
}

##########################################################################
#### rectInside rect1 rect2
#
# Test if rect1 is inside rect2.
#
proc rectInside {rect1 rect2} {
    assign x0 y0 x1 y1 $rect1
    assign a0 b0 a1 b1 $rect2
    
    return [expr $x0 >= $a0 && $y0 >= $b0 && $x1 <= $a1 && $y1 <= $b1]
}

##########################################################################
#### rectOverlap rect1 rect2
#
# Test if rect1 and rect2 overlap.
#
proc rectOverlap {rect1 rect2} {
    assign x0 y0 x1 y1 $rect1
    assign a0 b0 a1 b1 $rect2
    
    return [expr $a0 <= $x1 && $b0 <= $y1 && $a1 >= $x0 && $b1 >= $y0]
}

##########################################################################
#### rectUnion rect1 rect2
#
# Return the ``union'' of two rectangles: that is, the rectangle that
# contains both rectangles.
#
proc rectUnion {rect1 rect2} {
    assign x0 y0 x1 y1 $rect1
    assign a0 b0 a1 b1 $rect2
    
    return [list [min $x0 $a0] [min $y0 $b0] [max $x1 $a1] [max $y1 $b1]]
}

##########################################################################
#### rectAspect rect aspect ... aspect
#
# Return one or more aspects of a rectangle. Valid aspects are any
# of the eight compass points and "c" or "center."
#
proc rectAspect {rect args} {

    set result {}
    foreach aspect $args {
	switch -exact $aspect {
	    "n" {
		lappend result \
		    [expr ([lindex $rect 0] + [lindex $rect 2]) / 2]
		    [lindex $rect 1]
	    }
	    "s" {
		lappend result \
			[expr ([lindex $rect 0] + [lindex $rect 2]) / 2] \
			[lindex $rect 3]
	    }
	    "e" {
		lappend result \
			[lindex $rect 2] \
			[expr ([lindex $rect 1] + [lindex $rect 3]) / 2]
	    }
	    "w" {
		lappend result \
			[lindex $rect 0] \
			[expr ([lindex $rect 1] + [lindex $rect 3]) / 2]
	    }
	    "nw" {
		lappend result \
			[lindex $rect 0] \
			[lindex $rect 1]
	    }
	    "ne" {
		lappend result \
			[lindex $rect 2] \
			[lindex $rect 1]
	    }
	    "sw" {
		lappend result \
			[lindex $rect 0] \
			[lindex $rect 3]
	    }
	    "se" {
		lappend result \
			[lindex $rect 2] \
			[lindex $rect 3]
	    }
	    "c" -
	    "center" {
		lappend result \
			[expr ([lindex $rect 0] + [lindex $rect 2]) / 2] \
			[expr ([lindex $rect 1] + [lindex $rect 3]) / 2]
	    }
	}
    }
    return $result
}


##########################################################################
#
## Line, vertex, and polygon functions
#
# The TK canvas support multi-edge lines, specified as a series of
# points. The vertex functions assist in dealing with this kind of
# line. Each of these functions accepts a ``line,'' which is a flat
# list of x and y co-ordinates, and an index, which specifies which
# of its vertices to operate on.
#

##########################################################################
#### vertexAdd line index point
#
# Add a point to a vertex.
#
proc vertexAdd {line index point} {
    set m [expr $index*2]
    set n [expr $index*2+1]

    return [eval lreplace [list $line] $m $n \
	    [pointAdd [lrange $line $m $n] $point]]
}

##########################################################################
#### lineAspect
#
# Return one or more aspects of a line. Valid aspects are:
# <ul>
# <li>*first*: The first point of the line
# <li>*last*: The last point of the line
# <li>*vertex-*_n_: The _n_'th point of the line.
# </ul>
proc lineAspect {line args} {
    set result {}
    foreach aspect $args {
	set splut [split $aspect "-"]
	switch -exact [lindex $splut 0] {
	    "first" {
		lappend result [lindex $line 0]
		lappend result [lindex $line 1]
	    }
	    "last" {
		lappend result [expr [llength $line] - 2]
		lappend result [lindex $line end]
	    }
	    "vertex" {
		set n [lindex $splut 1]
		lappend result [lindex $line [expr 2*$n]]
		lappend result [lindex $line [expr 2*$n+1]]
	    }
	}
    }
    return $result
}

##########################################################################
#### lineReverse line
#
# Reverse the direction of a line. Basically just reverse
# a list of coordinates.
#
proc lineReverse {line} {
    set result {}
    while { [llength $line] >= 2 } {
	set a [lindex $line 0]
	set b [lindex $line 1]
	set line [lrange $line 2 end]
	set result [concat $a $b $result]
    }
    return $result
}

##########################################################################
#### lineFlip coords x
#
# Reflect a line about the specified x coordinate.
#
proc lineFlip {coords x0} {
    set result {}
    foreachpair x y $coords {
	set x [expr 2 * $x0 - $x]
	lappend result $x $y
    }
    return $result
}

##########################################################################
#### lineFlop coords y
#
# Reflect a line about the specified y coordinate.
#
proc lineFlop {coords y0} {
    set result {}
    foreachpair x y $coords {
	set y [expr 2 * $y0 - $y]
	lappend result $x $y
    }
    return $result
}

##########################################################################
#### lineRotate coords x y angle
#
# Rotate a line/polygon about the specified x and y coordinates, by the
# specified angle (in degrees).
#
proc lineRotate {coords x y angle} {
    set angle [radians $angle]
    set result {}

    while { [llength $coords] > 0 } {
	set a [lindex $coords 0]
	set b [lindex $coords 1]
	set coords [lreplace $coords 0 1]

	if { $a == $x && $y == $b } { ;# same point
	    lappend result $a $b
	} else {
	    set dx [expr $a - $x]
	    set dy [expr $b - $y]

	    set radius [expr sqrt($dx*$dx+$dy*$dy)]
	    set theta  [expr atan2($dy, $dx) + $angle]

	    set a [expr $x + $radius * cos($theta)]
	    set b [expr $y + $radius * sin($theta)]

	    lappend result $a $b
	}
    }

    return $result
}

##########################################################################
#### polyStretch distance coords [-angles var]
#
# Stretch a polygon. Requires that first and last coords be the same.
# The coordinates must proceed around the polygon in a clockwise
# direction to get the right sense of direction.
#
# The -angles options introduces a variable which is set to a list
# of the angles of each segment. The angles are calculated here
# anyway so this seemed like a good way to get them without having
# to calculate them again.
#
# NOTE: Because of Tk's upside-down co-ordinates, the angles will come
# out upside-down as well. Sorry 'bout that.
#
proc polyStretch {t args} {
    global pi

    set v [readopt angles args]
    if {$v != ""} {
	upvar $v angles
    }
    set coords [lindex $args 0]
    set angles {}
    set result {}
    
    set x0 [lindex $coords [expr [llength $coords] - 4]]
    set y0 [lindex $coords [expr [llength $coords] - 3]]

    behead x1 coords
    behead y1 coords

    set beta [expr atan2($y0-$y1, $x0-$x1)]

    while { [llength $coords] >= 2 } {
	set x2 [lindex $coords 0]
	set y2 [lindex $coords 1]
	set coords [lrange $coords 2 end]

	set alpha [expr atan2($y2-$y1, $x2-$x1)]
	set gamma [expr ($alpha - $beta) / 2.0]
	set theta [expr $alpha - $gamma]

	set delta [expr double($t) / sin($gamma)]

	set xt [expr $x1 + $delta * cos($theta)]
	set yt [expr $y1 + $delta * sin($theta)]

	lappend result $xt $yt
	lappend angles $alpha

	set x0 $x1
	set y0 $y1
	set x1 $x2
	set y1 $y2

	set beta [expr $alpha + $pi]
    }

    lappend result [lindex $result 0] [lindex $result 1]
    return $result
}

##########################################################################
#### polyBorders minRadians maxRadians coords
#
# Calculate coordinates needed to make polygon borders. It
# returns a list of three elements:
# <ul>
# <li> the coordinates of the inside of the polygon.
# <li> a list of lists of coordinates of polygons representing lit edges.
# <li> a list of lists of coordinates of polygons representing shaded edges.
# </ul>
#
# The minAngle and maxAngle arguments are the angles between which an
# edge is considered to be lit.
#    
proc polyBorders {minAngle maxAngle coords borderWidth} {

    set minRadians [radians $minAngle]
    set maxRadians [radians $maxAngle]

    set lit {}
    set shaded {}

    #
    # Get the coordinates of the two parallel polygons
    #
    set outer $coords
    set inner [polyStretch -$borderWidth $outer -angles angles]

    #
    # Rename to save for later
    #
    set inside $inner

    #
    # Calculate the lists of coordinates.
    #
    set a [lindex $angles 0]
    if { $a <= $minRadians && $a > $maxRadians } {
	set length 2
	set angles [lrange $angles 1 end]

	set a [lindex $angles 0]
	while { ($a <= $minRadians && $a > $maxRadians) \
		&& [llength $angles] > 0 } {
	    incr length 2
	    set angles [lrange $angles 1 end]
	    set a [lindex $angles 0]
	}

	lappend lit [concat \
		[ltake $outer [expr $length + 2]] \
		[lineReverse [ltake $inner [expr $length + 2]]] \
		[lindex $outer 0] \
		[lindex $outer 1]]

	set outer [ldrop $outer $length]
	set inner [ldrop $inner $length]
    }

    while { [llength $angles] > 0 } {
 	set length 2
	set angles [lrange $angles 1 end]
	set a [lindex $angles 0]
 	while { ($a > $minRadians || $a <= $maxRadians) \
		&& [llength $angles] > 0 } {
	    incr length 2
 	    set angles [lrange $angles 1 end]
	    set a [lindex $angles 0]
 	}

	lappend shaded [concat \
		[ltake $outer [expr $length + 2]] \
		[lineReverse [ltake $inner [expr $length + 2]]] \
		[lindex $outer 0] \
		[lindex $outer 1]]

 	if { [llength $angles] == 0 } {
 	    break
 	}

 	set outer [ldrop $outer $length]
 	set inner [ldrop $inner $length]

 	set length 2
	set angles [lrange $angles 1 end]
 	set a [lindex $angles 0]
 	while { ($a <= $minRadians && $a > $maxRadians) \
		&& [llength $angles] > 0 } {
 	    incr length 2
 	    set angles [lrange $angles 1 end]
 	    set a [lindex $angles 0]
 	}

	lappend lit [concat \
		[ltake $outer [expr $length + 2]] \
		[lineReverse [ltake $inner [expr $length + 2]]] \
		[lindex $outer 0] \
		[lindex $outer 1]]

 	set outer [ldrop $outer $length]
 	set inner [ldrop $inner $length]
    }

    #
    # Return the three-list of results
    #
    return [list $inside $lit $shaded]
}


##########################################################################
#
## Oval functions
#
# The TK canvas supports ellipses, or ``ovals,'' as it calls them,
# but provides no facilities for finding points on the oval. Here are
# some routines to help out with some of this stuff.
#
# The equation of an ellipse with horizontal radius _a_ and
# vertical radius _b_ is
#
#        2        2
#       x        y
#      ----  +  ----  =  1
#        2        2
#       a        b
#
# Let m = y/x (the tangent of the angle to (x,y).
#
# Acknowledgement: these functions were derived from Wan-teh Chang's
# finite-state-machine editor code.
#

## pi and friends
#
set pi    3.14159265358979
set twoPi [expr 2.0*$pi]
set piOverTwo [expr 0.5*$pi]
set piOverFour [expr 0.25*$pi]
set threePiOverTwo [expr 1.5*$pi]

##########################################################################
#### ovalFromRect
#
# Given a rectangle specifying the boundary of the ellipse,
# return a description in the form {x y a b}.
#
proc ovalFromRect {rect} {
    assign x0 y0 x1 y1 $rect
    set a [expr ($x1 - $x0) / 2.0]
    set x [expr $x1 - $a]
    set b [expr ($y1 - $y0) / 2.0]
    set y [expr $y1 - $b]

    return [list $x $y $a $b]
}

##########################################################################
#### ovalBoundary
#
# Given an oval and an angle, return the point on the circumference
# of the oval. The angle _must_ be between -pi and 2*pi.
#
proc ovalBoundary {oval angle} {
    global pi twoPi piOverTwo threePiOverTwo

    assign x y a b $oval
    
    #
    # First calculate m, checking for vertical angles
    #
    if { $angle > $pi } {
	set angle [expr $angle - $twoPi]
    }

    if {       $angle > [expr $piOverTwo - 0.01] \
	    && $angle < [expr $piOverTwo + 0.01] } {
	return [list $x [expr $y - $b]]
    } elseif { $angle > [expr $threePiOverTwo - 0.01] \
	    && $angle < [expr $threePiOverTwo + 0.01] } {
	return [list $x [expr $y + $b]]
    }
    set m [expr tan($angle)]

    #
    # Calculate point based on m
    #
    return [ovalM $oval $m [expr $angle >= 0.0]]
}

##########################################################################
#### ovalPoint
#
# Given an oval and a point, return the point on the circumference
# of the oval which intersects the line from its center to the
# point.
#
proc ovalPoint {oval point} {
    assign x y a b $oval
    assign p q $point

    #
    # First calculate m, checking for vertical angles
    #
    if { $p == $x } {
	if { $q < $y } {
	    return [list $x [expr $y - $b]]
	} else {
	    return [list $x [expr $y + $b]]
	}
    }
    set m [expr ($y - $q) / ($p - $x)]

    #
    # Calculate point based on m
    #
    return [ovalM $oval $m [expr $q < $y]]
}

##########################################################################
#### ovalM
#
# Given an oval and the tangent from its center,
# return the point on the circumference of the oval.
# The positive argument specifies whether the angle
# is positive or negative.
#
proc ovalM {oval m positive} {
    assign x y a b $oval

    set offset [expr $a*$b / sqrt($b*$b + $m*$m*$a*$a)]
    set offset [expr ($m < 0) ^ ! $positive ? -$offset : $offset]

    set pointx [expr $x + $offset]
    set pointy [expr $y - $m * $offset]

    return [list $pointx $pointy]
}
